import os
from typing import Any, Optional

from testcontainers.core.waiting_utils import wait_for_logs

from tomodachi_testcontainers.containers.common import DockerContainer
from tomodachi_testcontainers.utils import AWSClientConfig


class MotoContainer(DockerContainer):
    def __init__(
        self,
        image: str = "motoserver/moto:latest",
        internal_port: int = 5000,
        edge_port: int = 5000,
        region_name: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(image, **kwargs)
        self.internal_port = internal_port
        self.edge_port = edge_port

        self.region_name = region_name or os.environ.get("AWS_DEFAULT_REGION") or "eu-west-1"
        self.aws_access_key_id = os.environ.get("AWS_ACCESS_KEY_ID") or "testing"  # nosec: B105
        self.aws_secret_access_key = os.environ.get("AWS_SECRET_ACCESS_KEY") or "testing"  # nosec: B105

        self.with_bind_ports(self.internal_port, self.edge_port)
        self.with_env("AWS_DEFAULT_REGION", self.region_name)
        self.with_env("AWS_ACCESS_KEY_ID", self.aws_access_key_id)
        self.with_env("AWS_SECRET_ACCESS_KEY", self.aws_secret_access_key)

    def get_internal_url(self) -> str:
        ip = self.get_container_internal_ip()
        return f"http://{ip}:{self.internal_port}"

    def get_external_url(self) -> str:
        host = self.get_container_host_ip()
        return f"http://{host}:{self.edge_port}"

    def get_aws_client_config(self) -> AWSClientConfig:
        return AWSClientConfig(
            region_name=self.region_name,
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            endpoint_url=self.get_external_url(),
        )

    def start(self, timeout: float = 10) -> "MotoContainer":
        super().start()
        wait_for_logs(self, "Running on all addresses", timeout=timeout)
        return self

    def reset(self) -> None:
        self.exec(["sh", "-c", "curl", "-X", "POST", f"http://localhost:{self.internal_port}/moto-api/reset"])



from socket import socket
from typing import TypedDict


class AWSClientConfig(TypedDict):
    region_name: str
    aws_access_key_id: str
    aws_secret_access_key: str
    endpoint_url: str


def get_available_port() -> int:
    with socket() as sock:
        sock.bind(("", 0))
        return int(sock.getsockname()[1])



import os
import pathlib
from typing import Any, Dict, Iterator, Optional, Tuple, cast

import testcontainers.core.container
from docker.errors import ImageNotFound
from docker.models.images import Image as DockerImage
from testcontainers.core.docker_client import DockerClient


class DockerContainer(testcontainers.core.container.DockerContainer):
    def __init__(self, *args: Any, network: Optional[str] = None, **kwargs: Any) -> None:
        self.network = network or os.getenv("TESTCONTAINER_DOCKER_NETWORK", "bridge")
        super().__init__(*args, **kwargs, network=self.network)

    def get_container_internal_ip(self) -> str:
        container = self.get_docker_client().get_container(self.get_wrapped_container().id)
        return container["NetworkSettings"]["Networks"][self.network]["IPAddress"]


class EphemeralDockerImage:
    def __init__(self, dockerfile: pathlib.Path, docker_client_kw: Optional[Dict] = None) -> None:
        self.client = DockerClient(**(docker_client_kw or {}))
        self.image = self.build_image(dockerfile)

    def __enter__(self) -> DockerImage:
        return self.image

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.remove_image()

    def build_image(self, path: pathlib.Path) -> DockerImage:
        if path.is_dir():
            dockerfile_dir = path
            dockerfile_name = ""
        else:
            dockerfile_dir = path.parent
            dockerfile_name = path.name

        image, _ = cast(
            Tuple[DockerImage, Iterator],
            self.client.client.images.build(
                path=str(dockerfile_dir),
                dockerfile=str(dockerfile_name),
                forcerm=True,
            ),
        )
        return image

    def remove_image(self) -> None:
        self.client.client.images.remove(image=str(self.image.id))


def get_docker_image(image_id: str, docker_client_kw: Optional[Dict] = None) -> Optional[DockerImage]:
    client = DockerClient(**(docker_client_kw or {}))
    if not image_id:
        return None
    try:
        return cast(DockerImage, client.client.images.get(image_id))
    except ImageNotFound:
        return None



import os
from typing import AsyncGenerator, Generator

import asyncssh
import pytest
import pytest_asyncio

from tomodachi_testcontainers.containers import SFTPContainer
from tomodachi_testcontainers.utils import get_available_port


@pytest.fixture(scope="session")
def sftp_container() -> Generator[SFTPContainer, None, None]:
    image = os.environ.get("SFTP_TESTCONTAINER_IMAGE_ID", "atmoz/sftp:latest")
    with SFTPContainer(image=image, edge_port=get_available_port()) as container:
        yield container


@pytest_asyncio.fixture()
async def userpass_sftp_client(sftp_container: SFTPContainer) -> AsyncGenerator[asyncssh.SFTPClient, None]:
    conn_details = sftp_container.get_external_conn_details()
    async with asyncssh.connect(
        host=conn_details.host,
        port=conn_details.port,
        username="userpass",
        password="pass",  # nosec: B106
        known_hosts=sftp_container.get_known_hosts(),
    ) as ssh_conn:
        async with ssh_conn.start_sftp_client() as sftp_client:
            yield sftp_client


@pytest_asyncio.fixture()
async def userssh_sftp_client(sftp_container: SFTPContainer) -> AsyncGenerator[asyncssh.SFTPClient, None]:
    conn_details = sftp_container.get_external_conn_details()
    async with asyncssh.connect(
        host=conn_details.host,
        port=conn_details.port,
        username="userssh",
        client_keys=sftp_container.authorized_private_key,
        known_hosts=sftp_container.get_known_hosts(),
    ) as ssh_conn:
        async with ssh_conn.start_sftp_client() as sftp_client:
            yield sftp_client



import os
import pathlib
from typing import Generator

import pytest
from docker.models.images import Image as DockerImage

from tomodachi_testcontainers.containers import EphemeralDockerImage, get_docker_image


@pytest.fixture(scope="session")
def tomodachi_image() -> Generator[DockerImage, None, None]:
    if image := get_docker_image(os.environ.get("TOMODACHI_TESTCONTAINER_IMAGE_ID", "")):
        yield image
    else:
        dockerfile = pathlib.Path(os.environ.get("TOMODACHI_TESTCONTAINER_DOCKERFILE_PATH", "."))
        with EphemeralDockerImage(dockerfile) as image:
            yield image



from datetime import datetime, timedelta, timezone

UUID4_PATTERN = r"[0-9a-f]{8}-?[0-9a-f]{4}-?4[0-9a-f]{3}-?[89ab][0-9a-f]{3}-?[0-9a-f]{12}"

DEFAULT_RANGE = timedelta(seconds=10)


def assert_datetime_within_range(value: datetime, range: timedelta = DEFAULT_RANGE) -> None:
    current_datetime = datetime.utcnow().replace(tzinfo=timezone.utc)
    start_datetime = current_datetime - range
    end_datetime = current_datetime + range
    assert start_datetime <= value <= end_datetime  # nosec: B101



import json
from typing import Any, Dict, List, Protocol, Type, TypeVar, Union

from types_aiobotocore_sns import SNSClient
from types_aiobotocore_sqs import SQSClient

MessageType = TypeVar("MessageType")


class TomodachiSNSSQSEnvelope(Protocol):
    @classmethod
    async def build_message(cls: "TomodachiSNSSQSEnvelope", service: Any, topic: str, data: Any, **kwargs: Any) -> str:
        ...

    @classmethod
    async def parse_message(cls: "TomodachiSNSSQSEnvelope", payload: str, **kwargs: Any) -> Union[dict, tuple]:
        ...


async def subscribe_to(sns_client: SNSClient, sqs_client: SQSClient, topic: str, queue: str) -> None:
    create_topic_response = await sns_client.create_topic(Name=topic)
    topic_arn = create_topic_response["TopicArn"]

    create_queue_response = await sqs_client.create_queue(QueueName=queue)
    queue_url = create_queue_response["QueueUrl"]

    get_queue_attributes_response = await sqs_client.get_queue_attributes(
        QueueUrl=queue_url, AttributeNames=["QueueArn"]
    )
    queue_arn = get_queue_attributes_response["Attributes"]["QueueArn"]

    await sns_client.subscribe(TopicArn=topic_arn, Protocol="sqs", Endpoint=queue_arn)


async def receive(
    sqs_client: SQSClient,
    queue_name: str,
    envelope: TomodachiSNSSQSEnvelope,
    message_type: Type[MessageType],
    max_messages: int = 10,
) -> List[MessageType]:
    get_queue_url_response = await sqs_client.get_queue_url(QueueName=queue_name)
    queue_url = get_queue_url_response["QueueUrl"]

    received_messages_response = await sqs_client.receive_message(QueueUrl=queue_url, MaxNumberOfMessages=max_messages)

    parsed_messages: List[MessageType] = []
    for message in received_messages_response["Messages"]:
        parsed_message = await envelope.parse_message(json.loads(message["Body"])["Message"])
        parsed_messages.append(parsed_message[0]["data"])
    return parsed_messages


async def publish(sns_client: SNSClient, topic: str, data: Dict, envelope: TomodachiSNSSQSEnvelope) -> None:
    message = await envelope.build_message(service={}, topic=topic, data=data)

    create_topic_response = await sns_client.create_topic(Name=topic)
    topic_arn = create_topic_response["TopicArn"]

    await sns_client.publish(TopicArn=topic_arn, Message=message)



import contextlib

from tomodachi_testcontainers.containers.common import EphemeralDockerImage, get_docker_image
from tomodachi_testcontainers.containers.localstack import LocalStackContainer
from tomodachi_testcontainers.containers.moto import MotoContainer
from tomodachi_testcontainers.containers.tomodachi import TomodachiContainer

with contextlib.suppress(ModuleNotFoundError):
    from tomodachi_testcontainers.containers.sftp import SFTPContainer


__all__ = [
    "EphemeralDockerImage",
    "LocalStackContainer",
    "MotoContainer",
    "SFTPContainer",
    "TomodachiContainer",
    "get_docker_image",
]



from typing import Any, NamedTuple

import asyncssh
from testcontainers.core.waiting_utils import wait_for_logs

from tomodachi_testcontainers.containers.common import DockerContainer

ConnectionDetails = NamedTuple("ConnectionDetails", [("host", str), ("port", int)])


class SFTPContainer(DockerContainer):
    def __init__(
        self,
        image: str = "atmoz/sftp:latest",
        internal_port: int = 22,
        edge_port: int = 2222,
        **kwargs: Any,
    ) -> None:
        super().__init__(image, **kwargs)
        self.internal_port = internal_port
        self.edge_port = edge_port
        self.with_bind_ports(self.internal_port, self.edge_port)

        self.with_command("userpass:pass:1001 userssh::1002")

        self.authorized_private_key = asyncssh.generate_private_key("ssh-ed25519")
        self.authorized_public_key = self.authorized_private_key.export_public_key().decode()

    def get_internal_conn_details(self) -> ConnectionDetails:
        host = self.get_container_internal_ip()
        return ConnectionDetails(host=host, port=self.internal_port)

    def get_external_conn_details(self) -> ConnectionDetails:
        host = self.get_container_host_ip()
        return ConnectionDetails(host=host, port=self.edge_port)

    def get_host_public_key(self) -> str:
        _, output = self.exec("bash -c 'cat /etc/ssh/ssh_host_ed25519_key.pub'")
        return bytes(output).decode().strip()

    def get_internal_known_host(self) -> str:
        internal = self.get_internal_conn_details()
        public_key = self.get_host_public_key()
        return f"{internal.host} {public_key}"

    def get_external_known_host(self) -> str:
        external = self.get_external_conn_details()
        public_key = self.get_host_public_key()
        return f"{external.host} {public_key}"

    def get_known_hosts(self) -> asyncssh.SSHKnownHosts:
        known_hosts = asyncssh.SSHKnownHosts()
        for known_host in [self.get_internal_known_host(), self.get_external_known_host()]:
            known_hosts.load(known_host)
        return known_hosts

    def add_authorized_key(self, username: str, uid: str, gid: str, public_key: str) -> None:
        self.exec(f"bash -c 'mkdir -p /home/{username}/.ssh'")
        self.exec(f"bash -c 'chmod 700 /home/{username}/.ssh'")
        self.exec(f"bash -c 'touch /home/{username}/.ssh/authorized_keys'")
        self.exec(f"bash -c 'chmod 600 /home/{username}/.ssh/authorized_keys'")
        self.exec(f"bash -c 'echo \"{public_key}\" >> /home/{username}/.ssh/authorized_keys'")
        self.exec(f"bash -c 'chown -R {uid}:{gid} /home/{username}/.ssh'")

    def start(self, timeout: int = 10) -> "SFTPContainer":
        super().start()
        wait_for_logs(self, r"Server listening on", timeout=timeout)

        self.exec("bash -c 'mkdir /home/userpass/upload && chown -R 1001:1001 /home/userpass/upload'")
        self.exec("bash -c 'mkdir /home/userssh/upload && chown -R 1002:1002 /home/userssh/upload'")

        self.add_authorized_key(username="userssh", uid="1002", gid="1002", public_key=self.authorized_public_key)

        return self



from typing import Any

from testcontainers.core.waiting_utils import wait_for_logs

from tomodachi_testcontainers.containers.common import DockerContainer


class TomodachiContainer(DockerContainer):
    def __init__(self, image: str, internal_port: int = 9700, edge_port: int = 9700, **kwargs: Any) -> None:
        super().__init__(image, **kwargs)
        self.internal_port = internal_port
        self.edge_port = edge_port
        self.with_bind_ports(internal_port, edge_port)

    def get_internal_url(self) -> str:
        ip = self.get_container_internal_ip()
        return f"http://{ip}:{self.internal_port}"

    def get_external_url(self) -> str:
        host = self.get_container_host_ip()
        return f"http://{host}:{self.edge_port}"

    def start(self, timeout: float = 10) -> "TomodachiContainer":
        super().start()
        wait_for_logs(self, "Started service", timeout=timeout)
        return self



import os
from typing import AsyncGenerator, Generator

import pytest
import pytest_asyncio
from aiobotocore.session import get_session
from types_aiobotocore_dynamodb import DynamoDBClient
from types_aiobotocore_s3 import S3Client
from types_aiobotocore_sns import SNSClient
from types_aiobotocore_sqs import SQSClient
from types_aiobotocore_ssm import SSMClient

from tomodachi_testcontainers.containers import LocalStackContainer
from tomodachi_testcontainers.utils import get_available_port


@pytest.fixture(scope="session")
def localstack_container() -> Generator[LocalStackContainer, None, None]:
    image = os.environ.get("LOCALSTACK_TESTCONTAINER_IMAGE_ID", "localstack/localstack:2.1")
    with LocalStackContainer(image=image, edge_port=get_available_port()) as container:
        yield container


@pytest.fixture()
def _restart_localstack_container(localstack_container: LocalStackContainer) -> Generator[None, None, None]:
    yield
    localstack_container.restart()


@pytest_asyncio.fixture()
async def localstack_sns_client(localstack_container: LocalStackContainer) -> AsyncGenerator[SNSClient, None]:
    async with get_session().create_client("sns", **localstack_container.get_aws_client_config()) as c:
        yield c


@pytest_asyncio.fixture()
async def localstack_sqs_client(localstack_container: LocalStackContainer) -> AsyncGenerator[SQSClient, None]:
    async with get_session().create_client("sqs", **localstack_container.get_aws_client_config()) as c:
        yield c


@pytest_asyncio.fixture()
async def localstack_s3_client(localstack_container: LocalStackContainer) -> AsyncGenerator[S3Client, None]:
    async with get_session().create_client("s3", **localstack_container.get_aws_client_config()) as c:
        yield c


@pytest_asyncio.fixture()
async def localstack_dynamodb_client(localstack_container: LocalStackContainer) -> AsyncGenerator[DynamoDBClient, None]:
    async with get_session().create_client("dynamodb", **localstack_container.get_aws_client_config()) as c:
        yield c


@pytest_asyncio.fixture()
async def localstack_ssm_client(localstack_container: LocalStackContainer) -> AsyncGenerator[SSMClient, None]:
    async with get_session().create_client("ssm", **localstack_container.get_aws_client_config()) as c:
        yield c



import contextlib

import pytest

from tomodachi_testcontainers.pytest.localstack_fixtures import (
    _restart_localstack_container,
    localstack_container,
    localstack_dynamodb_client,
    localstack_s3_client,
    localstack_sns_client,
    localstack_sqs_client,
    localstack_ssm_client,
)
from tomodachi_testcontainers.pytest.moto_fixtures import (
    _reset_moto_container,
    moto_container,
    moto_dynamodb_client,
    moto_s3_client,
    moto_sns_client,
    moto_sqs_client,
    moto_ssm_client,
)
from tomodachi_testcontainers.pytest.tomodachi_fixtures import tomodachi_image

with contextlib.suppress(ModuleNotFoundError):
    from tomodachi_testcontainers.pytest.sftp_fixtures import sftp_container, userpass_sftp_client, userssh_sftp_client


pytest.register_assert_rewrite("tomodachi_testcontainers.pytest.assertions")


__all__ = [
    "_reset_moto_container",
    "_restart_localstack_container",
    "localstack_container",
    "localstack_dynamodb_client",
    "localstack_s3_client",
    "localstack_sns_client",
    "localstack_sqs_client",
    "localstack_ssm_client",
    "moto_container",
    "moto_dynamodb_client",
    "moto_s3_client",
    "moto_sns_client",
    "moto_sqs_client",
    "moto_ssm_client",
    "sftp_container",
    "tomodachi_image",
    "userpass_sftp_client",
    "userssh_sftp_client",
]
